//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#include <boost/numeric/ublas/io.hpp>
#include <boost/program_options.hpp>

#include "lwpr.hpp"

#include "rf_diag.hpp"
#include "rf_full.hpp"

LWPR::LWPR() : vm(boost::program_options::variables_map())
{

    // Default parameter vaules taken from paper and other implementations
    w_gen = 0.2;
    w_update = 0.001;
    w_predict = 0.001;
    diag = true;
    
}

LWPR::LWPR(const boost::program_options::variables_map& v) : vm(v){
    w_gen = vm["w_gen"].as<Types::RParam>();
    w_update = vm["w_update"].as<Types::RParam>();
    w_predict = vm["w_predict"].as<Types::RParam>();
    diag = vm["diag"].as<bool>();
}

LWPR::~LWPR()
{
    for(RFList::iterator it = receptiveFields.begin(); it != receptiveFields.end(); ++it)
        delete *it;
}


void LWPR::learn(Types::Input x, Types::Output y)
{
    // Taken From Table 5 in PAPER

    // For Each Receptive field
    //         1) Compute the activation of x at each receptive field
    //         2) Update the Distance Matrix
    //         3) Update the Local Model (projections and regression)
    //         4) Check if more projections are needed
    //
    // If no RF was activated by more than w_gen create a new RF with
    // R=2, c=x, and D=d_def*I;
    
    Types::RValue maxActivation = 0.0;
    Types::RValue tempActivaction = 0.0;
    RFList::iterator maxActivationIndex = receptiveFields.begin();

    for(RFList::iterator it = receptiveFields.begin(); it != receptiveFields.end(); ++it){
        tempActivaction = (*it)->getActivation(x);

        if(tempActivaction > w_update)
            (*it)->learn(x,y,tempActivaction);

        if(tempActivaction > maxActivation){
            maxActivation = tempActivaction;
            maxActivationIndex = it;
        }
    }

    if(maxActivation < w_gen){
        // make the new RF have similar shape to maximally actiavted rf if it is trustworthy.
        if((maxActivation > (0.1 * w_gen)) && (*maxActivationIndex)->trustworthy()){
            if(diag){
                receptiveFields.push_back(new RFDiag(*(RFDiag*)(*maxActivationIndex),x,vm));
            }
            else{
                receptiveFields.push_back(new RFFull(*(RFFull*)(*maxActivationIndex),x,vm));
            }
        }
        else{
            if(diag){
                receptiveFields.push_back(new RFDiag(x,vm));
            }
            else{
                receptiveFields.push_back(new RFFull(x,vm));
            }
        }
    }
}

Types::OutputT LWPR::predict(Types::Input x) const
{
    // Taken from Equation 3.2 in PAPER
    
    Types::RValue sumActivations = 0.0;
    Types::RValue tempActivaction = 0.0;
    Types::RValue sumPredictions = 0.0;

    for(RFList::const_iterator it = receptiveFields.begin(); it != receptiveFields.end(); ++it){
        tempActivaction = (*it)->getActivation(x);

        if(tempActivaction > w_predict && (*it)->trustworthy()){
            sumPredictions += tempActivaction * (*it)->predict(x);
            sumActivations += tempActivaction;
        }
    }

    // really shouldn't compare to 0 should do < 2*eps 
    return (sumActivations==0)?0:sumPredictions/sumActivations;
}


std::ostream& operator<<(std::ostream& out, const LWPR& lwpr)
{
    int i=0;

    // print out the receptive field centers and distance matrices
    for(LWPR::RFList::const_iterator it = lwpr.receptiveFields.begin(); it != lwpr.receptiveFields.end(); ++it){

        // this is terrible, got to be a better way
        if(lwpr.diag)
            out << *((RFDiag*)*it) << std::endl;
        else
            out << *((RFFull*)*it) << std::endl;

    }
}

//std::istream& operator>>(std::istream& in, LWPR& lwpr)
//{
//
//}
