//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#ifndef LWPR_HPP
#define LWPR_HPP

#include <list>
#include <iostream>
#include <fstream>

#include <boost/program_options.hpp>

#include "types.hpp"
#include "rf.hpp"

/**
 * Implements the high level interface to the LWPR algorithm.  The basic interface consists of
 * funtions for learning and predicting as well as getting error estimates.
 *
 */

class LWPR {

    /**
     * \name Algorithm Paramters
     *
     * This group of variables control the behavior of the algorithm.  All paramters are
     * defined in the paper (see PAPER).
     */

    //@{
    
    /**
     * w_gen controls the creation of new receptive fields.  If the activation of all RFs at a point is 
     * less than w_gen a new RF is created.
     */
    Types::RParam w_gen;    

    /**
     * w_update determines cut off for updating RFs.  If an RFs activation is greater than w_update it 
     * gets updated.
     */
    Types::RParam w_update;

    /**
     * w_predict determines if an RF is used in prediction.  Use an RF in the prediction if its 
     * activation is greater than w_predict.
     */
    Types::RParam w_predict;

    /**
     * diag determines whether a full or diagonal distance matrix is used.
     */
    bool diag;

    //@}

    /** 
     * Type of list used to store receptive fields 
     */
    typedef std::list<RF*> RFList;

    /**
     * A List of pointers to Receptive Fields for this lwpr module
     */
    LWPR::RFList receptiveFields;

    /**
     * reference to options map
     */
    const boost::program_options::variables_map& vm;

public:

    /** Default constructor for LWPR algorithm */
    LWPR();

    /**
     * This constructor takes a variables map and reads the necessary paramters for 
     * the class.  I'll probably make this dependent on a compiler flag since it 
     * requires an external shared library.
     */
    LWPR(const boost::program_options::variables_map& vm);

    /** Destructor.  Needs to free up memory associtated with RFs. */
    ~LWPR();

    
    /**
     * Update the LWPR model given a single data point.
     *
     * \param x the input data
     * \param y the value to regress on
     *
     */
    void learn(Types::Input x, Types::Output y);

    /**
     * Make a prediction for some test point.  The prediction is computed as in Equation 3.2
     * \f$ \hat y = \frac{\sum w_k \hat y_k}{\sum w_k} \f$
     *
     * \param x the input data to predict on
     *
     */
    Types::OutputT predict(Types::Input x) const;

    /**
     * Print the model to a stream in simple format
     */
    //friend std::istream& operator<<(std::istream& in, LWPR& lwpr);
    friend std::ostream& operator<<(std::ostream& out, const LWPR& lwpr);

};

#endif
