//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <iostream>
#include <iomanip>

//#include <boost/numeric/bindings/atlas/cblas.hpp>
//namespace bindings = boost::numeric::bindings::atlas;

#include "model.hpp"

Model::Model() {

};

Model::Model(int input_dim) : N(input_dim),R(2),
                                                            beta_0(0), 
                                                            x_0(N),
                                                            MSE(R),
                                                            data_count(R), derivatives_ok(R),
                                                            a_zz(R),a_zres(R),a_xz(N,R),
                                                            y_target(R),e_cv(R),
                                                            u(N,R),p(N,R),
                                                            x_res(N,R+1),beta(R),z(R),
                                                            x_temp(N), u_temp(N,R)
{

    u.clear();
    p.clear();

    u_temp.clear();

    // initialize projections to identity matrix
    u(0,0) = 1;
    p(0,0) = 1;

    if(N>1){
        u(1,1) = 1;
        p(1,1) = 1;
    }

    lambda = .999;
    phi = 0.5;

    x_0.clear();
    beta.clear();
    MSE.clear();

    // might move this into initilizer list
    data_count = ublas::scalar_vector<Types::RValue>(R,1e-10);
    derivatives_ok = ublas::zero_vector<Types::RValue>(R);
    a_zz = ublas::scalar_vector<Types::RValue>(R,1e-10);

    a_zres.clear();
    a_xz.clear();
}

Model::Model(int input_dim, const boost::program_options::variables_map& vm) : N(input_dim),R(2),
                                                            beta_0(0), 
                                                            x_0(N),
                                                            MSE(R),
                                                            data_count(R), derivatives_ok(R),
                                                            a_zz(R),a_zres(R),a_xz(N,R),
                                                            y_target(R),e_cv(R),
                                                            u(N,R),p(N,R),
                                                            x_res(N,R+1),beta(R),z(R),
                                                            x_temp(N), u_temp(N,R)
{

    u.clear();
    p.clear();

    u_temp.clear();

    // initialize projections to identity matrix
    u(0,0) = 1;
    p(0,0) = 1;

    if(N>1){
        u(1,1) = 1;
        p(1,1) = 1;
    }

    lambda = vm["lambda"].as<Types::RParam>();
    phi = vm["phi"].as<Types::RParam>();

    x_0.clear();
    beta.clear();

    MSE.clear();

    // might move this into initilizer list
    data_count = ublas::scalar_vector<Types::RValue>(R,1e-10);
    derivatives_ok = ublas::zero_vector<Types::RValue>(R);
    a_zz = ublas::scalar_vector<Types::RValue>(R,1e-10);

    a_zres.clear();
    a_xz.clear();
}


void Model::learn(Types::Input x, Types::Output y, Types::RValue w, Types::RValue W, Types::RValue W_old){
    x_0 = (lambda * W_old * x_0 + w * x)/W;
    beta_0 = (lambda * W_old * beta_0 + w * y)/W;

    // update current prediction error
    column(x_res,0) = x - x_0;
    y_hat = beta_0;

    for(int r=0; r<R; ++r){
        z(r) = prec_inner_prod(column(x_res,r),column(u,r));
        //z(r) = bindings::dot(column(x_res,r),column(u,r));

        y_hat += beta(r) * z(r);
        e_cv(r) = y-y_hat;
        
        column(x_res,r+1) = column(x_res,r) - z(r) * column(p,r);
        MSE(r) = lambda * MSE(r) + w * (y-y_hat) * (y-y_hat);
    }
    
    y_target(0) = y-beta_0;
    ublas::project(y_target, ublas::range(1,R)).assign(ublas::project(e_cv,ublas::range(0,R-1)));


    // update the local model
    res = y - beta_0;
    for(int r=0; r<R; ++r){
        a_zz(r) = lambda * a_zz(r) + w * z(r) * z(r);

        //a_zres(r) = lambda * a_zres(r) + w * z(r) * res;
        a_zres(r) = lambda * a_zres(r) + w * z(r) * y_target(r);

        beta(r) = a_zres(r) / a_zz(r);
        
        column(a_xz,r) = lambda * column(a_xz,r) + w * column(x_res,r) * z(r);

        // normalize the projections here rather than in the use above
        Types::RParam lambda_slow = 1 - (1-lambda)/10;

        //column(u_temp,r) = lambda_slow * column(u_temp,r) + w * res*column(x_res,r);
        column(u_temp,r) = lambda_slow * column(u_temp,r) + w * y_target(r)*column(x_res,r);
        
        Types::RValue norm_coeff = ublas::norm_2(column(u_temp,r));
        if(norm_coeff > 0)
            column(u,r) = column(u_temp,r)/norm_coeff;

        column(p,r) = column(a_xz,r) / a_zz(r);

        res -= z(r) * beta(r);
    }

    e = res;

    // update amount of data seen
    data_count = lambda * data_count + ublas::scalar_vector<Types::RValue>(R,1.0);

}

bool Model::trustworthy() const
{
    bool trust = true;
    
    for(int r=0; r<R; ++r)
        trust = trust && (data_count(r) > 2*N);    

    return trust;
}

Types::Input Model::check_derivatives() const
{
    for(int r=0; r<R; ++r)
        derivatives_ok(r) = (data_count(r) > 0.1/(1-lambda));

    return derivatives_ok;
}

bool Model::updateNumProjections()
{
    if( (R < N) && 
            (MSE(R-1)/MSE(R-2) <= phi) &&
            data_count(R-1)/data_count(0) > .9 &&
            data_count(R-1)*(1-lambda) > 0.5
            ){
        
        std::cerr << "****************** Adding Projection: " << x_0 << " ******************" << std::endl;
        
        // add projection
        R += 1;

        // add new projections to u and p all 1s
        u.resize(N,R);
        p.resize(N,R);
        a_xz.resize(N,R);

        for(int i=0; i<N; ++i){
            u(i,R-1) = (i == R-1);
            p(i,R-1) = (i == R-1);
            a_xz(i,R-1) = 0;
        }

        a_zz.resize(R);
        a_zz(R-1) = 1e-10;
        
        a_zres.resize(R);

        MSE.resize(R);
        MSE(R-1) = 0;

        e_cv.resize(R);
        y_target.resize(R);

        u_temp.resize(N,R);
        for (int i=0;i<N;++i)
            u_temp(i,R-1) = 0;

        data_count.resize(R);
        data_count(R-1) = 1e-10;
        derivatives_ok.resize(R);
        derivatives_ok(R-1) = 0;

        x_res.resize(N,R+1);
        beta.resize(R);
        z.resize(R);
        
        return true;
    }
    else
        return false;
}

Types::OutputT Model::predict(Types::Input x) const {
    Types::OutputT y = beta_0;

    Types::vector &x_local = const_cast<Types::vector&>(x_temp);

    x_local = x - x_0;

    for(int r=0; r<R; ++r){
        Types::RValue s = prec_inner_prod(column(u,r),x_local);
        //Types::RValue s = bindings::dot(column(u,r),x_local);
        y += beta(r) * s;
        x_local -= s * column(p,r);
    }

    return y;
        
}

void Model::project(Types::Input x, Types::vector& z) const {
    z.resize(x.size());
    z = x - x_0;

    for(int r=0; r<R; ++r){
        Types::RValue s = prec_inner_prod(column(u,r),z);
        //Types::RValue s = bindings::dot(column(u,r),z);
        z -= s * column(p,r);
    }
}

std::ostream& operator<<(std::ostream& out, const Model& m)
{
    out << m.u << std::endl;
}

