//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#ifndef MODEL_HPP
#define MODEL_HPP

#include <boost/program_options.hpp>

#include "types.hpp"

/** 
 * A locally weighted partial least squares implementation.  The weights come from a receptive 
 * field, RF.  The models basic interface is learn(input, output) and predict(input).
 */

class Model {

    /** the number of projections. */
    int R;

    /** the dimension of the input */
    int N;

    /**
     * \name Algorithm Parameters
     *
     * This group of variables control the behavior of the algorithm.  All paramters are
     * defined in the paper (see PAPER).
     */

    //@{
    
    /**
     * A forgetting factor for sufficient statistics.  This parameter must have a value in [0,1].  
     * It may be updated over time.
     */
    Types::RParam lambda;

    /**
     * A threshold for adding new projections.  If \f$ \frac{MSE_t+1}{MSE_t} > \phi \f$ then add a projection.
     * Must take a value in [0,1]
     */
    Types::RParam phi;

    //@}

    

    /**
     * \name Sufficient Statistics
     *
     * These variables hold the sufficient statistics for updating the RF and local model.
     *
     */

    //@{
    
    /** Weighted Input History **/
    Types::vector x_0;

    /** Weighted Output History **/
    Types::RValue beta_0;

    /** MSE memory */
    Types::vector MSE;

    Types::vector a_zz, a_zres;

    Types::Matrix a_xz;


    /** amount of data seen for projection dimension */
    Types::vector data_count;

    /** checks for derivatives being valid, uses data_count */
    mutable Types::vector derivatives_ok;

    //@}
    
    /** Projection Directions */
    Types::Matrix u, p, u_temp;

    /** error variables */
    Types::RValue e;

    /** cross validation error for each dimension */
    Types::vector e_cv;

    /** store the errors for each dimension */
    Types::vector y_target;

    // temp variables
    Types::Matrix x_res;
    Types::vector beta, z;
    Types::RValue y_hat, res;

    Types::vector x_temp;


public:
    /**
     * The default constructor.
     */
    Model();

    /**
     * The constructor.
     */
    Model(int input_dim);

    /**
     * A constructor taking a program_options::variables_map for getting paramter values.
     */
    Model(int input_dim, const boost::program_options::variables_map& vm);


    /**
     * Update the model given a weighted data point and output.
     *
     * \param x the input vector
     * \param y the true output
     * \param w the current activation (from RF)
     * \param W the activation history (from RF)
     * \param W_old the old activation history (from RF)
     */
    void learn(Types::Input x, Types::Output y, Types::RValue w, Types::RValue W, Types::RValue W_old);

    /**
     * Check whether enough data has been seen to trust this model
     */
    bool trustworthy() const;

    /** Check whether derivatives with respect to projections are ok */
    Types::Input check_derivatives() const;

    /** 
     * Check if a new projection is needed.  If it is add it.  This requires param \f$\phi\f$.
     */
    bool updateNumProjections();

    /**
     * Project an input to local low dimensional space.
     *
     * \param x the input vector
     * \param z the projected vector
     *
     * \return void
     */
    void project(Types::Input x, Types::vector& z) const;

    /** 
     * Predict a value from a given input.
     *
     * \param x the input vector
     *
     * \return The predicted value at x
     */
    Types::OutputT predict(Types::Input x) const;

    /** return last estimate of e_cv */
    Types::RValue get_e_cv() const { return e_cv(e_cv.size()-1); };

    /** return last estimate of e */
    Types::RValue get_e() const { return e; };

    /** return last estimate of MSE */
    Types::RValue get_MSE() const { return MSE(R-1); };

    /** get last estimate of a_zz */
    Types::Input get_a_zz() const { return a_zz; };

    /** get last projected x */
    Types::Input get_z() const { return z; };

    /**
     * Print the model to a stream in simple format
     */
    friend std::ostream& operator<<(std::ostream& out, const Model& m);


};

#endif 
