//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#ifndef RF_HPP
#define RF_HPP

#include <math.h>

#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/program_options.hpp>

#include <fstream>

#include "types.hpp"
#include "model.hpp"

/**
 * The RF class represents a receptive field.  This class handles updates of the distance matrix for
 * the receptive field and hold the local model associated with the field.
 */

class RF {
protected:

    /**
     * \name Algorithm Parameters
     *
     * This group of variables control the behavior of the algorithm.  All paramters are
     * defined in the paper (see PAPER).
     */

    //@{
    
    /**
     * d_def is the initial value for diagonal of distance matrix in RF.
     */
    Types::RParam d_def;

    /**
     * A forgetting factor for sufficient statistics.  This parameter must have a value in [0,1].  
     * It may be updated over time.
     */
    Types::RParam lambda;

    /**
     * A weight on the distance matrix penalty.
     */
    Types::RParam gamma;

    /**
     * alpha_init is the initial learning rate for all distance elements
     */
    Types::RParam alpha_init;

    /** 
     * determine if meta updates are done
     */
    bool meta;

    /**
     * meta_learning rate
     */
    Types::RValue meta_alpha;

    //@}
    

    /**
     * \name RF Descriptors
     *
     * These variables describe the receptive field.  They include the center, distance matrix, and
     * local regression model
     */

    //@{
    
    /** The center point of the RF */
    Types::vector center;

    /** the receptive fields local model */
    Model localModel;
    
    //@}

    
    /**
     * \name Sufficient Statistics
     *
     * These variables hold the sufficient statistics for updating the RF and local model.
     *
     */

    //@{
    
    /** Activation history **/
    Types::RValue W;

    Types::vector a_H, a_G;
    Types::RValue a_E;

    /** confidence intervals ss */
    Types::RValue a_pk, sigma;

    Types::RValue e_2;

    //@}

    /** store the projection of x */
    Types::vector z;

    // temp variables, this isn't terribly safe, but its saves having to reallocate
    // the damn things every time they're needed
    mutable Types::vector temp2, q, H_temp;

public:

    /** Default constructor for the RF */
    RF();

    /**
     * A constructor taking only a center point for the RF.
     *
     * \param c An input vector corresponding to the center of the RF.
     */
    RF(Types::Input c);

    /**
     * A constructor taking a center point and a variables_map.
     */
    RF(Types::Input c, const boost::program_options::variables_map& vm);


    /**
     * Return the activation for a test point.  This currently uses ublas vectro expressions, but
     * should be changed to use blas calls once everything is working correctly.
     *
     * \param x the input vector
     */
    virtual Types::RValue getActivation(Types::Input x) const = 0;

    /**
     * Return the predicted value for an unput x.
     *
     * \param x the input vector
     */
    inline Types::OutputT predict(Types::Input x) const { return localModel.predict(x); };

    /** 
     * return the predictive variance for a point
     *
     * \param x the input point
     */
    Types::OutputT confidence(Types::Input x) const;


    /**
     * Check whether the RF is trustworthy
     *
     * \see Model::trustworthy()
     */
    inline bool trustworthy() const { return localModel.trustworthy(); };


    /**
     * Update the receptive field and have local model update itself.
     * 
     * Update the distance matrix as in Equation 3.6 and Table 4
     * \f$ M^{n+1} = M^n - \alpha \frac{\partial J}{\partial M} \f$
     *
     * \param x the input vector
     * \param y the actual output
     *
     */
    virtual Types::RValue learn(Types::Input x, Types::Output y, Types::RValueP w) = 0;

    /**
     * Print the receptive field to a stream in simple format
     */
    friend std::ostream& operator<<(std::ostream& out, const RF& rf);

};

#endif
