//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_expression.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <algorithm>
#include <iomanip>

#include <math.h>

#include "rf_diag.hpp"
#include "my_expression.hpp"

RFDiag::RFDiag()
{

}

RFDiag::RFDiag(Types::Input c) : 
    RF(c),
    D(c.size()), 
    M(c.size()), 
    alpha(c.size()),
    dwdM(c.size()),
    dJ2dM(c.size()),
    dJdM(c.size()),
    meta_h(c.size()),
    b(c.size()),
    dJ2_2dM_2(c.size()), 
    dw_2dM_2(c.size()), 
    aux(c.size())
{
    D.assign(d_def*ublas::scalar_vector<Types::RValue>(c.size()));
    M.assign(sqrt(d_def)*ublas::scalar_vector<Types::RValue>(c.size()));
    alpha.assign(ublas::scalar_vector<Types::RParam>(c.size(),alpha_init));

    meta_h.clear();
    b.assign(ublas::scalar_vector<Types::RParam>(c.size(),std::log(alpha_init+1e-10)));
}

RFDiag::RFDiag(Types::Input c, const boost::program_options::variables_map& vm) : 
    RF(c,vm),
    D(c.size()), 
    M(c.size()), 
    alpha(c.size()),
    dwdM(c.size()),
    dJ2dM(c.size()),
    dJdM(c.size()),
    meta_h(c.size()),
    b(c.size()),
    dJ2_2dM_2(c.size()), 
    dw_2dM_2(c.size()), 
    aux(c.size())
{
    std::cerr << "Adding Receptive Field" << std::endl;
    D.assign(d_def*ublas::scalar_vector<Types::RValue>(c.size()));
    M.assign(sqrt(d_def)*ublas::scalar_vector<Types::RValue>(c.size()));
    alpha.assign(ublas::scalar_vector<Types::RParam>(c.size(),alpha_init));

    meta_h.clear();
    b.assign(ublas::scalar_vector<Types::RParam>(c.size(),std::log(alpha_init+1e-10)));
}

RFDiag::RFDiag(const RFDiag& rf, Types::Input c, const boost::program_options::variables_map& vm) : 
    RF(c, vm),
    D(c.size()), 
    M(c.size()), 
    alpha(c.size()),
    dwdM(c.size()),
    dJ2dM(c.size()),
    dJdM(c.size()),
    meta_h(c.size()),
    b(c.size()),
    dJ2_2dM_2(c.size()), 
    dw_2dM_2(c.size()), 
    aux(c.size())
{
    alpha.assign(rf.getAlpha());
    D.assign(rf.getD());
    M.assign(rf.getM());

    meta_h.clear();
    b.assign(ublas::scalar_vector<Types::RParam>(c.size(),std::log(alpha_init+1e-10)));
}

Types::RValue RFDiag::learn(Types::Input x, Types::Output y, Types::RValueP w){
    // get the current activation
//    const Types::RValue w = getActivation(x);

    // update some sufficient statistics
    const Types::RValue W_old = W;
    W = lambda * W + w;

    //update the local model
    localModel.learn(x,y,w,W,W_old);

    // stuff from local model, this relies on learn being called first, perhaps I should rework this
    const Types::RValue e_cv = localModel.get_e_cv();
    const Types::RValue e = localModel.get_e();
    Types::Input z = localModel.get_z();


    // Update the distance matrix as in Equation 3.6 and Table 4
    // \f$ M^{n+1} = M^n - \alpha \frac{\partial J}{\partial M} \f$
    
    // do some strange stuff to help with numerical stability
    Types::Input derivative = localModel.check_derivatives();

    if(derivative(0) != 0){    // update is ok

        // update squared error trackers
        e_2 = lambda * e_2 + w*e*e;

        // this seems odd but paper says update it here, before use
        // update a_E
        a_E = lambda * a_E + w*e_cv*e_cv;
    
        Types::RValue transient_multiplier = std::min(1.0, pow( e_2 / (a_E + 1e-10) , 4));

        // compute \f$ \sum_{i=1}^M \frac{\partial J_1}{\partial w} \f$
        q = element_div(element_prod(z,derivative),localModel.get_a_zz());

        // sufficient statistic for confidence intervals
        a_pk = lambda*a_pk + w*w*prec_inner_prod(q,z);
        
        const Types::RValue dJ1dw = (e_cv*e_cv)/W - 2*e/W * prec_inner_prod(q,a_H) - 2/W*prec_inner_prod(element_prod(q,q),a_G) - a_E/(W*W);

        temp2 = x - center;

        // code for diagonal M
        for(int k=0; k<dwdM.size(); ++k){
          const Types::RValue dDdM = 2*M(k);
          dwdM(k) = -0.5*w*temp2(k)*temp2(k)*dDdM;
          dJ2dM(k) = 2 * gamma / center.size() * D(k) * dDdM;

            // declare dJ2_2dM_2 and dw_2dM_2
            if(meta){
                const Types::RValue dDdM_2 = dDdM*dDdM;
                dJ2_2dM_2(k) = 2*gamma/center.size()*(2*D(k)+dDdM_2);
                dw_2dM_2(k) = 2*dwdM(k)*dwdM(k)/w - w*temp2(k)*temp2(k);
            }
        }

        dJdM = dJ1dw * dwdM + (w/W)*dJ2dM;
    
        // compute h
        const Types::RValue h = w * prec_inner_prod(z,q);

        // declare (dbouel)dJ1_2dw_2, (vector)dJ_2dM_2, vector)aux, (vector)b, (vector) meta_h,
        // exp(matrix) function
        if(meta){
            const Types::RValue dJ1_2dw_2 = -(e_cv*e_cv)/(W*W) 
                          + 2*e*e*h/(W*w) 
                                    + a_E/(W*W*W) 
                                    - 2/W*((-(e/W) - 2*e*prec_inner_prod(q,z))*prec_inner_prod(q,a_H))
                                    - 1/(W*W)*( (e_cv*e_cv) - 2*e*prec_inner_prod(q,a_H) );

            dJ_2dM_2 = (dw_2dM_2*dJ1dw + element_prod(dwdM,dwdM)*dJ1_2dw_2) + w/W*dJ2_2dM_2;

            aux = meta_alpha * transient_multiplier * element_prod(dJdM,meta_h);

            for(int i=0; i<aux.size(); ++i)
                if(fabs(aux(i)) > 0.1)
                    aux(i) = 0.1 * ((aux(i)>0)?1:-1);

            b -= aux;

            for(int i=0; i<b.size(); ++i)
                if(fabs(b(i)) > 10)
                    b(i) = 10 * ((b(i)>0)?1:-1);

            // probably need to define this
//            alpha = exp(b);
//            alpha = ublas::apply_to_all<functor::exp<Types::RValue> >( b );
            alpha = (ublas::apply_to_all<ublas::scalar_exp<Types::RValue> >( b ));

            aux = ublas::scalar_vector<Types::RValue>(alpha.size(),1.0) - (element_prod(alpha,dJ_2dM_2)*transient_multiplier);

            for(int i=0; i<aux.size(); ++i)
                if(aux(i) < 0.0)
                    aux(i) = 0;

            meta_h = element_prod(meta_h,aux) - element_prod(alpha,dJdM)*transient_multiplier;

        }

        // update a_H, a_G
        H_temp =  lambda * a_H + (w*e_cv*z*transient_multiplier)/(1-h);
        a_H = element_prod(derivative, H_temp) + element_prod(ublas::scalar_vector<Types::RValue>(derivative.size(),1.0)-derivative,a_H);
        H_temp = lambda * a_G + (w*w*(e_cv*e_cv)*element_prod(z,z)*transient_multiplier)/(1-h);
        a_G = element_prod(derivative, H_temp) + element_prod(ublas::scalar_vector<Types::RValue>(derivative.size(),1.0)-derivative,a_G);

        Types::RValue maxM = 0.0;
        for(int i=0; i<M.size(); ++i)
            if(fabs(M(i)) > maxM)
                maxM = fabs(M(i));

        for(int i=0; i<M.size(); ++i)
            if(alpha(i) * dJdM(i) * transient_multiplier > .1*maxM)
                alpha(i) /= 2;

        // code for diagonal M, also need to change getActivation()
        M -= element_prod(alpha,dJdM) * transient_multiplier;
        D = element_prod(M,M);

    }


    // Check if projection needs to be added, update local ss if projection added
    if(localModel.updateNumProjections()){
        int R = a_H.size()+1;
        a_H.resize(R);
        a_H(R-1) = 0;
        a_G.resize(R);
        a_G(R-1) = 0;
        H_temp.resize(R);
    }

}

std::ostream& operator<<(std::ostream& out, const RFDiag& rf)
{
    for(int i=0; i<rf.center.size(); ++i)
        out << rf.center(i) << " ";

    for(int i=0; i<rf.D.size(); ++i)
        for(int j=0; j<rf.D.size(); ++j)
            if(i==j)
                out << rf.D(0) << " ";
            else 
                out << 0.0 << " ";
    out << std::endl;

}
