//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#ifndef RF_DIAG_HPP
#define RF_DIAG_HPP

#include <math.h>

#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/program_options.hpp>

#include <fstream>

#include "types.hpp"
#include "rf.hpp"


/**
 * The RFDiag class represents a receptive field with a diagonal matrix.  This class 
 * handles updates of the distance matrix for
 * the receptive field and hold the local model associated with the field.
 */

class RFDiag : public RF {

    /**
     * \name Meta Learning Sufficient Statistics 
     *
     * This group of variables control the behavior of the algorithm.  All paramters are
     * defined in the paper (see PAPER).
     */

    //@{
    
    /**
     * Learning rate for the distance matrix updates.
     */
    Types::vector alpha;

    /**
     * Meta Update Params
     */
    Types::vector meta_h,b;

    //@}
    

    /**
     * \name RF Descriptors
     *
     * These variables describe the receptive field.  They include the center, distance matrix, and
     * local regression model
     */

    //@{
    
    /** the distance matrix of the RF */
    Types::vector D;

    /** The Cholesky decomposition of D.  This is used for updating the distance matrix. */
    Types::vector M;

    //@}


    // temp variables
    Types::vector dwdM, dJ2dM, dJdM;
    
    // meta learning temp vars
    Types::vector dJ2_2dM_2, aux, dw_2dM_2, dJ_2dM_2;

public:

    /** Default constructor for the RF */
    RFDiag();

    /**
     * A constructor taking only a center point for the RF.
     *
     * \param c An input vector corresponding to the center of the RF.
     */
    RFDiag(Types::Input c);

    /**
     * A constructor taking a center point and a variables_map.
     */
    RFDiag(Types::Input c, const boost::program_options::variables_map& vm);

    /**
     * a copy constructor, sort of.  Just copy vm, alpha, D, M.
     */
    RFDiag(const RFDiag&, Types::Input c, const boost::program_options::variables_map& vm);

    /**
     * Return the activation for a test point.  This currently uses ublas vectro expressions, but
     * should be changed to use blas calls once everything is working correctly.
     *
     * \param x the input vector
     */
    virtual Types::RValue getActivation(Types::Input x) const
    { 
        //const_cast<Types::vector&>(temp2) = element_prod(M,(x-center));
        Types::RValue act = 0.0;

        for(int i=0; i<M.size(); ++i)
            act += M(i)*(x(i)-center(i))*M(i)*(x(i)-center(i));

        return exp(-0.5 * act);
        //return exp(-0.5 * ublas::prec_inner_prod(temp2,temp2)); 
    };


    /**
     * Update the receptive field and have local model update itself.
     * 
     * Update the distance matrix as in Equation 3.6 and Table 4
     * \f$ M^{n+1} = M^n - \alpha \frac{\partial J}{\partial M} \f$
     *
     * \param x the input vector
     * \param y the actual output
     *
     */
    virtual Types::RValue learn(Types::Input x, Types::Output y, Types::RValueP w);
    
    /**
     * print the center and distance matrix on a line
     */
    friend std::ostream& operator<<(std::ostream& out, const RFDiag& rf);

    /**
     * stupid get functions for copy constructor.
     */
    Types::Input getAlpha() const { return alpha; };
    Types::Input getD() const { return D; };
    Types::Input getM() const { return M; };
    Types::Input getCenter() const { return center; };


};

#endif
