//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_expression.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <algorithm>
#include <iomanip>

#include <math.h>

#include "rf_full.hpp"
#include "my_expression.hpp"

RFFull::RFFull()
{

}

RFFull::RFFull(Types::Input c) : 
    RF(c),
    D(c.size(),c.size()), 
    M(c.size(),c.size()), 
    alpha(c.size(),c.size()),
    dwdM(c.size(),c.size()),
    dJ2dM(c.size(),c.size()),
    dJdM(c.size(),c.size()),
    meta_h(c.size(),c.size()),
    b(c.size(),c.size()),
    dJ2_2dM_2(c.size(),c.size()), 
    dw_2dM_2(c.size(),c.size()), 
    aux(c.size(),c.size())
{
    D.assign(d_def*ublas::identity_matrix<Types::RValue>(c.size()));
    M.assign(sqrt(d_def)*ublas::identity_matrix<Types::RValue>(c.size()));
    alpha.assign(ublas::scalar_matrix<Types::RParam>(c.size(),c.size(),alpha_init));

    dwdM.clear();
    dJ2dM.clear();
    dJdM.clear();
    dJ2_2dM_2.clear();
    dw_2dM_2.clear();
    aux.clear();

    meta_h.clear();
    b.assign(ublas::scalar_matrix<Types::RParam>(c.size(),c.size(),std::log(alpha_init+1e-10)));
}

RFFull::RFFull(Types::Input c, const boost::program_options::variables_map& vm) : 
    RF(c,vm),
    D(c.size(),c.size()), 
    M(c.size(),c.size()), 
    alpha(c.size(),c.size()),
    dwdM(c.size(),c.size()),
    dJ2dM(c.size(),c.size()),
    dJdM(c.size(),c.size()),
    meta_h(c.size(),c.size()),
    b(c.size(),c.size()),
    dJ2_2dM_2(c.size(),c.size()), 
    dw_2dM_2(c.size(),c.size()), 
    aux(c.size(),c.size())
{
    D.assign(d_def*ublas::identity_matrix<Types::RValue>(c.size()));
    M.assign(sqrt(d_def)*ublas::identity_matrix<Types::RValue>(c.size()));

    alpha.assign(ublas::scalar_matrix<Types::RParam>(c.size(),c.size(),alpha_init));

    dwdM.clear();
    dJ2dM.clear();
    dJdM.clear();
    dJ2_2dM_2.clear();
    dw_2dM_2.clear();
    aux.clear();

    meta_h.clear();
    b.assign(ublas::scalar_matrix<Types::RParam>(c.size(),c.size(),std::log(alpha_init+1e-10)));
}

RFFull::RFFull(const RFFull& rf, Types::Input c, const boost::program_options::variables_map& vm) : 
    RF(c, vm),
    D(c.size(),c.size()), 
    M(c.size(),c.size()), 
    alpha(c.size(),c.size()),
    dwdM(c.size(),c.size()),
    dJ2dM(c.size(),c.size()),
    dJdM(c.size(),c.size()),
    meta_h(c.size(),c.size()),
    b(c.size(),c.size()),
    dJ2_2dM_2(c.size(),c.size()), 
    dw_2dM_2(c.size(),c.size()), 
    aux(c.size(),c.size())
{
    std::cerr << "Adding Receptive Field" << std::endl;
    alpha.assign(rf.getAlpha());
    D.assign(rf.getD());
    M.assign(rf.getM());

    dwdM.clear();
    dJ2dM.clear();
    dJdM.clear();
    dJ2_2dM_2.clear();
    dw_2dM_2.clear();
    aux.clear();

    meta_h.clear();
    b.assign(ublas::scalar_matrix<Types::RParam>(c.size(),c.size(),std::log(alpha_init+1e-10)));
}


Types::RValue RFFull::learn(Types::Input x, Types::Output y, Types::RValueP w) {
    // get the current activation
//    const Types::RValue w = getActivation(x);

    // update some sufficient statistics
    const Types::RValue W_old = W;
    W = lambda * W + w;

    //update the local model
    localModel.learn(x,y,w,W,W_old);

    // stuff from local model, this relies on learn being called first, perhaps I should rework this
    const Types::RValue e_cv = localModel.get_e_cv();
    const Types::RValue e = localModel.get_e();
    Types::Input z = localModel.get_z();

    // Update the distance matrix as in Equation 3.6 and Table 4
    // \f$ M^{n+1} = M^n - \alpha \frac{\partial J}{\partial M} \f$
    
    // do some strange stuff to help with numerical stability
    Types::Input derivative = localModel.check_derivatives();

    if(derivative(0) != 0){    // update is ok

        // update squared error trackers
        e_2 = lambda * e_2 + w*e*e;
        
        // this seems odd but paper says update it here, before use
        // update a_E
        a_E = lambda * a_E + w*e_cv*e_cv;

        Types::RValue transient_multiplier = std::min(1.0, pow( e_2 / (a_E + 1e-10) , 4));

        // compute \f$ \sum_{i=1}^M \frac{\partial J_1}{\partial w} \f$
        q = element_div(element_prod(z,derivative),localModel.get_a_zz());

        // sufficient statistic for confidence intervals
        a_pk = lambda*a_pk + w*w*prec_inner_prod(q,z);

        const Types::RValue dJ1dw = (e_cv*e_cv)/W - 2*e/W * prec_inner_prod(q,a_H) - 2/W*prec_inner_prod(element_prod(q,q),a_G) - a_E/(W*W);

        temp2 = x - center;

        dwdM.clear();

        // M is upper triangular
        for(int k=0; k<dwdM.size1(); ++k)
            for(int l=k; l<dwdM.size2(); ++l){
                Types::RValue sum_aux = 0;
                Types::RValue dDdM_2 = 0;
                
                // compute dDdM
                for(int i=0; i<dwdM.size1(); ++i){
                    sum_aux += D(i,l)*M(k,i)*2;
                    dwdM(k,l) += temp2(i)*temp2(l)*M(k,i)*((i==l)?2:1);

                    if(meta)
                        dDdM_2 += (M(k,i)*M(k,i)*((i==l)?2:1)) * 2;

                }
                
                dwdM(k,l) = -0.5 * w * dwdM(k,l);
                dJ2dM(k,l) = 2 * gamma / center.size() * sum_aux;

                if(meta)
                {
                    dJ2_2dM_2(k,l) = 2*gamma/center.size()*(2*D(l,l) + dDdM_2);
                    dw_2dM_2(k,l) = dwdM(k,l)*dwdM(k,l)/w - w*temp2(k)*temp2(k);
                }

            }

        dJdM = dJ1dw * dwdM + (w/W)*dJ2dM;

        // compute h
        const Types::RValue h = w * prec_inner_prod(z,q);
    
        if(meta){
            const Types::RValue dJ1_2dw_2 = -(e_cv*e_cv)/(W*W)
                                  + 2*e*e*h/(W*w)
                                    + a_E/(W*W*W)
                                    - 2/W*((-(e/W) - 2*e*prec_inner_prod(q,z))*prec_inner_prod(q,a_H))
                                    - 1/(W*W)*( (e_cv*e_cv) - 2*e*prec_inner_prod(q,a_H) );

            dJ_2dM_2 = (dw_2dM_2*dJ1dw + element_prod(dwdM,dwdM)*dJ1_2dw_2) + w/W*dJ2_2dM_2;

            aux = meta_alpha * transient_multiplier * element_prod(dJdM,meta_h);

            for(int i=0; i<aux.size1(); ++i)
                for(int j=i; j<aux.size2(); ++j)
                    if(fabs(aux(i,j)) > 0.1)
                        aux(i,j) = 0.1 * ((aux(i,j)>0)?1:-1);

            b -= aux;

            for(int i=0; i<b.size1(); ++i)
                for(int j=i; j<b.size2(); ++j)
                    if(fabs(b(i,j)) > 10)
                        b(i,j) = 10 * ((b(i,j)>0)?1:-1);

            //alpha = exp(b);
            alpha = (ublas::apply_to_all<ublas::scalar_exp<Types::RValue> >( b ));

            aux = ublas::scalar_matrix<Types::RValue>(alpha.size1(),alpha.size2(),1.0) - (element_prod(alpha,dJ_2dM_2)*transient_multiplier);

            for(int i=0; i<aux.size1(); ++i)
                for(int j=i; j<aux.size2(); ++j)
                    if(aux(i,j) < 0.0)
                        aux(i,j) = 0;

            meta_h = element_prod(meta_h,aux) - element_prod(alpha,dJdM)*transient_multiplier;
        }
        
        // update a_H, a_G
        H_temp =  lambda * a_H + (w*e_cv*z*transient_multiplier)/(1-h);
        a_H = element_prod(derivative, H_temp) + element_prod(ublas::scalar_vector<Types::RValue>(derivative.size(),1.0)-derivative,a_H);
        H_temp = lambda * a_G + (w*w*(e_cv*e_cv)*element_prod(z,z)*transient_multiplier)/(1-h);
        a_G = element_prod(derivative, H_temp) + element_prod(ublas::scalar_vector<Types::RValue>(derivative.size(),1.0)-derivative,a_G);

        // reduce learning rate for elements of M where the update is large
        Types::RValue maxM = 0.0;
        for(int i=0; i<M.size1(); ++i)
            for(int j=i; j<M.size2(); ++j)
                if(fabs(M(i,j)) > maxM)
                    maxM = fabs(M(i,j));

        for(int i=0; i<alpha.size1(); ++i)
            for(int j=i; j<alpha.size2(); ++j)
                if(alpha(i,j) * dJdM(i,j) * transient_multiplier > .1*maxM)
                    alpha(i,j) /= 2;


        // update M and D
        M -= element_prod(alpha,dJdM) * transient_multiplier;
        D = prod(trans(M),M);

    }


    // Check if projection needs to be added, update local ss if projection added
    if(localModel.updateNumProjections()){
        int R = a_H.size()+1;
        a_H.resize(R);
        a_H(R-1) = 0;
        a_G.resize(R);
        a_G(R-1) = 0;
        H_temp.resize(R);
    }

}

std::ostream& operator<<(std::ostream& out, const RFFull& rf)
{
    out << rf.center(0) << " " << rf.center(1) << " ";
    out << rf.D(0,0) << " " << rf.D(0,1) << " " << rf.D(1,0) << " " << rf.D(1,1) << std::endl;
}
