//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#ifndef RF_FULL_HPP
#define RF_FULL_HPP

#include <math.h>

#include <boost/numeric/ublas/vector_expression.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/program_options.hpp>

#include <fstream>

#include "types.hpp"
#include "rf.hpp"

/**
 * The RFFull class represents a receptive field with a symmetric distance matrix.  
 * This class handles updates of the distance matrix for
 * the receptive field and hold the local model associated with the field.
 */

class RFFull : public RF {

    /**
     * \name Meta Learning Sufficient Statistics 
     *
     * This group of variables control the behavior of the algorithm.  All paramters are
     * defined in the paper (see PAPER).
     */

    //@{
    
    /**
     * Learning rate for the distance matrix updates.
     */
    //Types::UTMatrix alpha;
    Types::Matrix alpha;

    /**
     * Meta Update Params
     */
    //Types::UTMatrix meta_h,b;
    Types::Matrix meta_h,b;

    //@}
    

    /**
     * \name RF Descriptors
     *
     * These variables describe the receptive field.  They include the center, distance matrix, and
     * local regression model
     */

    //@{
    
    /** the distance matrix of the RF */
    //Types::SymMatrix D;
    Types::Matrix D;

    /** The Cholesky decomposition of D.  This is used for updating the distance matrix. */
    //Types::UTMatrix M;
    Types::Matrix M;
    
    //@}

    // temp variables
    //Types::UTMatrix dwdM, dJ2dM, dJdM;
    Types::Matrix dwdM, dJ2dM, dJdM;

    // meta learning temp vars
    //Types::UTMatrix dJ2_2dM_2, aux, dw_2dM_2, dJ_2dM_2;
    Types::Matrix dJ2_2dM_2, aux, dw_2dM_2, dJ_2dM_2;

public:

    /** Default constructor for the RF */
    RFFull();

    /**
     * A constructor taking only a center point for the RF.
     *
     * \param c An input vector corresponding to the center of the RF.
     */
    RFFull(Types::Input c);

    /**
     * A constructor taking a center point and a variables_map.
     */
    RFFull(Types::Input c, const boost::program_options::variables_map& vm);

    /**
     * a copy constructor, sort of.  Just copy vm, alpha, D, M.
     */
    RFFull(const RFFull&, Types::Input c, const boost::program_options::variables_map& vm);


    /**
     * Return the activation for a test point.  This currently uses ublas vectro expressions, but
     * should be changed to use blas calls once everything is working correctly.
     *
     * \param x the input vector
     */
    virtual Types::RValue getActivation(Types::Input x) const
    { 
        const_cast<Types::vector&>(temp2) = prod(M,(x-center));
        return exp(-0.5 * ublas::prec_inner_prod(temp2,temp2)); 
    };


    /**
     * Update the receptive field and have local model update itself.
     * 
     * Update the distance matrix as in Equation 3.6 and Table 4
     * \f$ M^{n+1} = M^n - \alpha \frac{\partial J}{\partial M} \f$
     *
     * \param x the input vector
     * \param y the actual output
     *
     */
    virtual Types::RValue learn(Types::Input x, Types::Output y, Types::RValueP w);

    /**
     * print the center and distance matrix on a line
     */
    friend std::ostream& operator<<(std::ostream& out, const RFFull& rf);

    /**
     * stupid get functions for copy constructor.
     */
    Types::Matrix getAlpha() const { return alpha; };
    Types::Matrix getD() const { return D; };
    Types::Matrix getM() const { return M; };
    Types::Input getCenter() const { return center; };

};

#endif
