//
//  Copyright (c) 2005-2007
//  James N Knight
//
//  Permission to use, copy, modify, distribute and sell this software
//  and its documentation for any purpose is hereby granted without fee,
//  provided that the above copyright notice appear in all copies and
//  that both that copyright notice and this permission notice appear
//  in supporting documentation.  The authors make no representations
//  about the suitability of this software for any purpose.
//  It is provided "as is" without express or implied warranty.
//
//  
// See http://homepages.inf.ed.ac.uk/svijayak/publications/vijayakumar-NeuCom2005.pdf
// for the original publication of this algorithm.



#include <fstream>
#include <iostream>
#include <string>
#include <sstream>

#include <vector>
#include <algorithm>

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/tuple/tuple.hpp>

#include <boost/program_options.hpp>

#include "lwpr.hpp"

using namespace std;
namespace po = boost::program_options;

int main(int argc, char* argv[]){

    ublas::vector<double> x(2);
    double y;
    double temp,temp2;

    std::vector<boost::tuple<ublas::vector<double>,double> > data;
    std::vector<boost::tuple<ublas::vector<double>,double> > data2;

    /**
     * Program Options
     */
    po::options_description generic("Generic Options");
    generic.add_options()
        ("help","Produce a help message.")
        ;

    po::options_description algorithm_parameters("Algorithm Paramters");
    algorithm_parameters.add_options()
        ("w_gen", po::value<Types::RParam>()->default_value(.2),"value for generating new RF")
        ("w_update", po::value<Types::RParam>()->default_value(.001),"activation for updating model params")
        ("w_predict", po::value<Types::RParam>()->default_value(.001),"activation for using model in prediction")
        ("lambda", po::value<Types::RParam>()->default_value(.999),"forgetting factor for sufficient statistics")
        ("gamma",  po::value<Types::RParam>()->default_value(1e-6),"penalty for distance metric update")
        ("d_def",  po::value<Types::RParam>()->default_value(25),"initial value for distance metric diagonal")
        ("phi",    po::value<Types::RParam>()->default_value(.5),"value for adding new projection")
        ("alpha",  po::value<Types::RParam>()->default_value(250),"initial learning rate for distance metric update")
        ("diag",   po::value<bool>()->default_value(true),"true if diag only distance matrix")
        ("meta",   po::value<bool>()->default_value(false),"true if meta updates are wanted")
        ("theta",   po::value<Types::RParam>()->default_value(250),"learning rate for second order updates")
        ;

    po::options_description experiment_parameters("Experiment Parameters");
    experiment_parameters.add_options()
        ("iterations", po::value<int>()->default_value(20),"number of iterations through the training data")
        ("train-data", po::value<std::string>(),"path to training data (required)")
        ("test-data",  po::value<std::string>(),"path to test data (required)")
        ("input-dim",  po::value<int>()->default_value(3),"number of input dimensions")
        ;

    po::options_description cmdline_options;
    cmdline_options.add(generic).add(algorithm_parameters).add(experiment_parameters);

    po::options_description config_file_options;
    config_file_options.add(generic).add(algorithm_parameters);

    po::positional_options_description p; 
    p.add("train-data",1);
    p.add("test-data",1);

    po::variables_map vm;
    store(po::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);

    //ifstream ifs("params.cfg");
    //store(parse_config_file(ifs, config_file_options), vm);
    //notify(vm);
    if (vm.count("help")) {
        std::cout << cmdline_options << "\n";
        return 0;
    }

    if (vm.count("train-data")){
        LWPR lwpr(vm);
    
        ifstream file(vm["train-data"].as<std::string>().c_str(),std::ios_base::in);

        ublas::vector<double> data_point(vm["input-dim"].as<int>());

        while( !file.eof() ){
            for(int i=0; i<vm["input-dim"].as<int>(); ++i)
                file >> data_point(i);
            file >> y;
            data.push_back(boost::make_tuple(data_point,y));    
        }
        file.close();
    
        file.open(vm["test-data"].as<std::string>().c_str(),std::ios_base::in);
        while( !file.eof() ){
            for(int i=0; i<vm["input-dim"].as<int>(); ++i)
                file >> data_point(i);
            file >> y;
            data2.push_back(boost::make_tuple(data_point,y));    
        }
        file.close();

        for(int i=0; i<vm["iterations"].as<int>(); ++i)
        {
            std::random_shuffle(data.begin(),data.end());
           
            // train model on all training data
            for(int count=0; count<data.size(); ++count){
                x = data[count].get<0>();
                y = data[count].get<1>();
    
                lwpr.learn(x,y);
            }

            // get error of trained model on training data
            double mse = 0;    
            for(int count=0; count<data.size(); ++count){
                x = data[count].get<0>();
                y = data[count].get<1>();
                const double pred = lwpr.predict(x);
                mse += (y-pred)*(y-pred);
            }
   
            // write out model data and prediction data
            std::string outfile1;// = "output/pred" + std::string(i);
            std::string outfile2;// = "output/model" + std::string(i);
    
            std::stringstream ss,ss2;
    
            ss << "output/pred" << i;
            ss >> outfile1;
            ss2 << "output/model" << i;
            ss2 >> outfile2;
    
            ofstream pred_out(outfile1.c_str(),std::ios_base::out);
            ofstream model_out(outfile2.c_str(),std::ios_base::out);
   
            // get error on test set
            double nmse = 0;
            for(int count=0; count<data2.size(); ++count){
                x = data2[count].get<0>();
                y = data2[count].get<1>();
    
                const double pred = lwpr.predict(x);
                pred_out << pred << std::endl;
                nmse += (y-pred)*(y-pred);
            }
    
            std::cout << "Iteration: " << i << " MSE_TRAIN: " << mse/data.size() << " MSE_TEST: " << nmse/data2.size() << std::endl;
    
            model_out << lwpr;
    
            pred_out.close();
            model_out.close();
    
        }
    }
    else{
        std::cout << cmdline_options << "\n";
        return 0;
    }

}
